import pytest
import yaml


@pytest.fixture
def cache_content():
    content = """
        access: eyJ0eXAiOiJKV1XPOCJhbGciOiJIUzI1NiJ9.eyJ0b2tlbl90eXBlIjoiYWNjZXNzIiwiZXhwIjoxNjEyNTM3OTk1LCJqdGkiOiI1ZjFlYjQ4ZTU2MTY0OTIzYjk5NzNlYWE3MTZmMDJiMyIsInVzZXJfaWQiOiJyb290In0.OlA9G9_ZIyoK8NViEDEoLFQxqwxLjwmzrio7wAgZHrs
        organisation_uuid: 8df668f217494cdfaff89c1a731f31cb
        refresh: eyJ0eXAiOiXPV1QiLCJhbGciOiJIUzI1NiJ9.eyJ0b2tlbl90eXBlIjoicmVmcmVzaCIsImV4cCI6MTYxMjYyNDM5NSwianRpIjoiYjVmOGUyZjgzMjhiNDdhZmFkN2U3MjJlMzQ1OGNhMjAiLCJ1c2VyX2lkIjoicm9vdCJ9.R1yi6i7uSMm9SbARzAsSORCCthlOhe8SvVLXbZFx9_U
        result_folder: /tmp/results
        scenario_folder: /tmp/scenarios
        username: test    
    """
    yield yaml.load(content, Loader=yaml.FullLoader)


@pytest.fixture
def scenario_constant_rain_content():
    content  = """
        meta:
            name: "constant rain"
            description: "adds a constant rain event with a duration of 5 minutes at the beginning of the simulation"
        
        scenario:
            simulation:
                threedimodel: "{{ threedimodel_id }}"
                organisation: "{{ organisation_uuid }}"
                name: "{{ simulation_name }}"
                start_datetime: "{{ datetime_now }}"
                duration: 3600
            steps:
                - constantrain:
                    offset: 0
                    duration: 3600
                    value: 0.00002
                    units: 'm/s'
                - action:
                    name: start
                    duration: 300
                    waitfor_timeout: 1800
                - action:
                    name: start
                    waitfor_timeout: 1800
                - waitforstatus:
                    name: 'finished'
                    timeout: 1800    
        """
    yield yaml.load(content, Loader=yaml.FullLoader)



from __future__ import annotations
from dataclasses import dataclass, field, asdict
from datetime import timedelta
from datetime import datetime
from enum import Enum
from pathlib import Path
import sys
from typing import Optional, List
from urllib.parse import urlparse
from functools import lru_cache
import json

import click
import jwt
import yaml
from rich.prompt import Prompt

from openapi_client import Configuration, ApiClient, ApiException
from threedi_api_client.threedi_api_client import ThreediApiClient
from threedi_api_client.aio.threedi_api_client import (
    ThreediApiClient as AsyncThreediApiClient,
)

from openapi_client.api.auth_api import AuthApi
from threedi_api_client.threedi_api_client import (
    is_token_usable,
    get_auth_token,
)

from threedi_cmd.console import console
from threedi_cmd.errors import ExitCodes

SCENARIO_DIR = Path(__file__).parent.parent / "scenarios"
REFRESH_TIME_DELTA = timedelta(hours=4).total_seconds()
EXPIRE_LEEWAY = -300


class EndpointOption(Enum):
    localhost = "http://localhost:8000/v3.0"
    staging = "https://api.staging.3di.live/v3.0"
    production = "https://api.3di.live/v3.0"


@lru_cache()
def get_settings(endpoint):
    return Settings(endpoint=endpoint)


@dataclass
class WebSocketSettings:
    api_base_url: str
    token: str
    host: str = field(init=False)
    proto: str = field(init=False)
    api_version: str = field(init=False)

    def __post_init__(self):
        parsed_url = urlparse(self.api_base_url)
        self.host = parsed_url.netloc
        self.proto = "wss" if parsed_url.scheme == "https" else "ws"
        self.api_version = parsed_url.path.lstrip("/")
        self.token = (
            self.token
            if self.token.startswith("Bearer")
            else f"Bearer {self.token}"
        )


@dataclass
class CachedConfig:
    username: str = ""
    access: str = ""
    refresh: str = ""
    organisation_uuid: str = ""
    result_folder: str = ""
    scenario_folder: str = ""

    @classmethod
    def load_from_file(cls, file: Path) -> CachedConfig:
        """
        Loads saved settings from 3di_config.yaml. See save method for available attributes.
        """
        try:
            with open(file, "r") as f:
                cached_data = yaml.load(f, Loader=yaml.FullLoader)
                return cls(**cached_data)
        except OSError:
            # settings file does not yet exist
            return cls()

    def save_to_file(self, file: Path) -> None:
        with open(file, "w") as f:
            yaml.dump(
                asdict(self),
                default_flow_style=False,
                stream=f,
            )


@dataclass
class ScenariosMeta:

    scenario_dir: Path = SCENARIO_DIR

    @property
    def scenarios(self) -> List:
        scenario_list = []
        scenario_files = self.scenario_dir.glob("*.yaml")
        for f in scenario_files:
            with f.open("r") as y:
                content = yaml.load_all(y, Loader=yaml.FullLoader)
                try:
                    meta = next(content)["meta"]
                    meta["file"] = f
                    scenario_list.append(meta)
                except KeyError:
                    raise AttributeError(
                        f"The scenario definition is missing the 'meta' section"
                    )
        return scenario_list


class Settings:
    """Settings that are saved (authentication) values between calls to the cli.

    The default settings are saved in the file '3di_config.yaml' which lives in
    the configuration folder of your operating system.
    """

    APP_NAME = "3di-scenario"
    CONFIG_FILE_NAME = "3di_config.yaml"

    def __init__(self, endpoint: str = ""):
        self._endpoint = EndpointOption[endpoint]
        self.cached_config = CachedConfig.load_from_file(self.config_file)
        self.refresh_api_key(self.configuration)
        self.websocket_settings = WebSocketSettings(
            api_base_url=self.endpoint, token=self.access
        )

    @property
    def endpoint(self) -> str:
        return self._endpoint.value

    @property
    def username(self) -> str:
        # print(f"prop username {self._username}")
        return self.cached_config.username

    @username.setter
    def username(self, username: str):
        print(f"setter username {username}")
        self.cached_config.username = username

    @property
    def access(self) -> str:
        return self.cached_config.access

    @access.setter
    def access(self, access: str):
        self.cached_config.access = access

    @property
    def refresh(self) -> str:
        return self.cached_config.refresh

    @refresh.setter
    def refresh(self, refresh: str):
        self.cached_config.refresh = refresh

    @property
    def organisation_uuid(self) -> str:
        return self.cached_config.organisation_uuid

    @organisation_uuid.setter
    def organisation_uuid(self, organisation_uuid: str):
        self.cached_config.organisation_uuid = organisation_uuid

    @property
    def result_folder(self) -> Path:
        return Path(self.cached_config.result_folder).resolve()

    @result_folder.setter
    def result_folder(self, result_folder: Path):
        if isinstance(result_folder, Path):
            result_folder = result_folder.resolve().as_posix()
        self.cached_config.result_folder = result_folder

    @property
    def scenario_folder(self) -> Path:
        return Path(self.cached_config.scenario_folder).resolve()

    @scenario_folder.setter
    def scenario_folder(self, scenario_folder: Path):
        if isinstance(scenario_folder, Path):
            scenario_folder = scenario_folder.resolve().as_posix()
        self.cached_config.scenario_folder = scenario_folder

    @property
    def app_dir(self) -> Path:
        _app_dir = Path(click.get_app_dir(app_name=self.APP_NAME))
        _app_dir.mkdir(parents=True, exist_ok=True)
        return _app_dir

    @property
    def config_file(self) -> Path:
        return self.app_dir / f"{self._endpoint.name}_{self.CONFIG_FILE_NAME}"

    @property
    def configuration(self) -> Configuration:
        configuration = Configuration(
            host=self.endpoint,
            username=f"{self.username}",
            api_key={
                "Authorization": f"{self.access}",
                "refresh": f"{self.refresh}",
            },
            api_key_prefix={"Authorization": "Bearer"},
        )
        configuration.refresh_api_key_hook = self.refresh_api_key
        return configuration

    @property
    def api_client(self) -> ThreediApiClient:
        client = ApiClient(self.configuration)
        return client

    @property
    def async_api_client(self) -> AsyncThreediApiClient:
        client = AsyncThreediApiClient(self.configuration)
        return client

    @property
    def scenarios(self) -> List:
        scenarios_meta = ScenariosMeta(self.scenario_folder)
        return scenarios_meta.scenarios

    def save(self) -> None:
        try:
            self.cached_config.save_to_file(self.config_file)
        except IOError as err:
            console.print(
                f"Saving cache to {self.config_file} failed with: {err}",
                style="error"
            )
            return

    @staticmethod
    def save_settings() -> None:
        ctx = click.get_current_context()
        ctx.obj.save()

    def do_refresh(self, token) -> bool:
        try:
            details = jwt.decode(
                token,
                options={"verify_signature": False},
                leeway=EXPIRE_LEEWAY,
            )
        except (jwt.exceptions.ExpiredSignatureError, jwt.exceptions.DecodeError):
            return True
        expiry_dt = datetime.fromtimestamp(details["exp"])
        sec_left = (expiry_dt - datetime.utcnow()).total_seconds()
        return sec_left <= REFRESH_TIME_DELTA

    def _call_auth(self, config: Configuration) -> None:
        api_client = ApiClient(Configuration(config.host))
        auth = AuthApi(api_client)
        try:
            token = auth.auth_refresh_token_create(
                {"refresh": config.api_key["refresh"]}
            )
            config.api_key = {"Authorization": token.access, "refresh": token.refresh}
            return
        except ApiException as err:
            if err.status == 403:  # forbidden
                console.print(
                    f"Too bad, we could not authenticate you. "
                    f"The server said: {err.reason}", style="error"
                )
                sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)
            self.credentials_prompt(config)

    def refresh_api_key(self, config: Configuration) -> None:
        """Refreshes the access key if its expired"""
        api_key = config.api_key.get("Authorization")
        if not self.do_refresh(api_key) and is_token_usable(api_key):
            self._call_auth(config)

        refresh_key = config.api_key.get("refresh")
        if is_token_usable(refresh_key):
            self._call_auth(config)
        else:
            self.credentials_prompt(config)

    def credentials_prompt(self, config: Configuration) -> None:
        console.rule(
            f":key: Authentication required for {config.host}",
            characters="*",
            style="gold3",
        )
        kwargs = {}
        try:
            if self.username:
                kwargs.update({"default": self.username})
        except AttributeError:
            pass
        username = Prompt.ask("Username", **kwargs)
        password = Prompt.ask("Password", password=True)

        try:
            token = get_auth_token(username, password, config.host)
        except ValueError as err:
            console.print(f":collision: {err}", style="error")
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)
        except ApiException as err:
            body = json.loads(err.body)["detail"]
            console.print(f"Could not log you in. The server responded with: {body}", style="error")
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)

        # Update the settings
        self.username = username
        self.access = token.access
        self.refresh = token.refresh
        config.api_key = {"Authorization": token.access, "refresh": token.refresh}



import asyncio
from datetime import datetime
from enum import Enum
import typer

from threedi_cmd.console import console
from threedi_cmd.models.monitor import ActiveSimulations
from threedi_cmd.commands.callbacks import default_callback
from threedi_cmd.commands.settings import (
    Settings,
    EndpointOption,
    get_settings,
)

active_sims_app = typer.Typer(callback=default_callback)


async def monitor_active_simulations(endpoint: str) -> None:
    """
    executes ActiveSimlations "run_monitor" task in the background
    """
    active_simulations = ActiveSimulations(endpoint)
    result = await asyncio.gather(
        active_simulations.run_monitor(), return_exceptions=True
    )
    if result:
        console.print(result, style="error")


@active_sims_app.command()
def simulations(
        ctx: typer.Context
):
    """
    Show currently running simulations
    """
    start_time = datetime.now().strftime("%Y-%m-%d %H:%M")
    try:
        console.print(f"[{start_time}] Starting active simulations worker")
        asyncio.run(monitor_active_simulations(ctx.obj._endpoint.name))
    except KeyboardInterrupt:
        pass
    finally:
        console.print(
            f":sparkles: Bye bye, hope to see you soon! :sparkles:",
            style="success",
        )


if __name__ == '__main__':
    active_sims_app()



from pathlib import Path

import pytest
import yaml
from threedi_cmd.commands.settings import CachedConfig
from .fixtures import cache_content


def test_load_cached_config_no_file():
    c = CachedConfig.load_from_file(Path(""))
    assert isinstance(c, CachedConfig)


def test_save_and_load(tmp_path, cache_content):
    c = CachedConfig(**cache_content)
    f = tmp_path / "test_config.yaml"
    c.save_to_file(f)
    assert f.exists() and f.is_file()
    rc = CachedConfig.load_from_file(f)
    assert rc == c




import pytest
import yaml
from threedi_cmd.commands.settings import ScenariosMeta
from threedi_cmd.commands.settings import CachedConfig

from .fixtures import cache_content, scenario_constant_rain_content


def test_scenario_meta(tmp_path, scenario_constant_rain_content):
    crf = tmp_path / "constant_rain.yaml"
    with open(crf, "w") as f:
        yaml.dump(scenario_constant_rain_content,
            default_flow_style=False,
            stream=f,
        )
    sm = ScenariosMeta(tmp_path)
    assert isinstance(sm.scenarios, list)
    assert len(sm.scenarios) == 1


def test_wrong_scenario_format(tmp_path, cache_content):
    c = CachedConfig(**cache_content)
    f = tmp_path / "test_config.yaml"
    c.save_to_file(f)
    sm = ScenariosMeta(tmp_path)

    with pytest.raises(AttributeError):
        # misses meta section
        sm.scenarios



import importlib
import pkgutil

from threedi_cmd.plugins.models import AppRegistry

PLUGIN_NAME_PREFIX = "threedi_cmd_"


def discover() -> AppRegistry:
    """find installed plugins"""

    discovered_plugins = {
        name: importlib.import_module(name)
        for finder, name, is_pkg
        in pkgutil.iter_modules()
        if name.startswith(PLUGIN_NAME_PREFIX)
    }
    registry = AppRegistry({})
    for plugin in discovered_plugins.values():
        registry.apps.update(plugin.registry.apps)
    return registry



from dataclasses import dataclass
from typing import Optional, Dict

import typer


@dataclass
class AppMeta:
    app: typer.Typer
    name: str
    help: str
    add_to: Optional[str] = ""


@dataclass
class AppRegistry:
    apps: Dict[str, AppMeta]



import asyncio
import sys
from datetime import datetime
from pathlib import Path

import click
import websockets
from openapi_client.api.simulations_api import SimulationsApi
from openapi_client.api.organisations_api import OrganisationsApi
from openapi_client.api.threedimodels_api import ThreedimodelsApi
from openapi_client import Configuration, ApiException
from rich.table import Table
from rich import box
from rich.prompt import Confirm
from rich.prompt import Prompt, IntPrompt
from rich.padding import Padding
from rich.panel import Panel
import typer

from threedi_cmd.commands.utils import download_files
from threedi_cmd.commands.scenarios import scenarios
from threedi_cmd.parser import ScenarioParser
from threedi_cmd.console import console
from threedi_cmd.commands.utils import PathPrompt
from threedi_cmd.errors import ExitCodes
from threedi_cmd.models.errors import ApiModelError
from threedi_cmd.models.scenario import ResolveError
from threedi_cmd.models.scenario import FailedStep
from threedi_cmd.errors import LoadScenarioError
from threedi_cmd.logger import get_logger
from threedi_cmd.commands.callbacks import default_callback

logger = get_logger("INFO")

DEFAULT_PAGER_SIZE = 20


app = typer.Typer()
api_app = typer.Typer(callback=default_callback)
app.add_typer(api_app, name="api")


@api_app.command()
def models(ctx: typer.Context):
    """List available threedimodels"""
    threedi_models_api = ThreedimodelsApi(ctx.obj.api_client)
    threedi_models = threedi_models_api.threedimodels_list()

    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column("Id", width=5)
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )
    table.add_column("Revision", justify="left")
    table.add_column("Inp success", justify="left")

    remaining = threedi_models.count
    limit = min(DEFAULT_PAGER_SIZE, remaining)

    while remaining == threedi_models.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        threedi_models = threedi_models_api.threedimodels_list(
            limit=limit, offset=remaining - limit
        )
        for i, model in enumerate(threedi_models.results):
            if model.inp_success is True:
                txt = f":heavy_check_mark: [bold green]{model.inp_success}[/bold green] "
            else:
                txt = f"[bold red]{model.inp_success}[/bold red]"
            table.add_row(
                f"{model.id}", f"{model.name}", f"{model.revision_hash}", txt
            )
        remaining -= limit
        limit = min(remaining, DEFAULT_PAGER_SIZE)
        console.print(table)


@api_app.command()
def organisations(
        ctx: typer.Context):
    """List available organisations"""
    organisations_api = OrganisationsApi(ctx.obj.api_client)
    organisations = organisations_api.organisations_list()
    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column("Unique Id", width=15)
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )

    remaining = organisations.count
    limit = min(DEFAULT_PAGER_SIZE, remaining)

    while remaining == organisations.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        organisations = organisations_api.organisations_list(
            limit=limit, offset=remaining - limit
        )
        for i, org in enumerate(organisations.results):
            table.add_row(f"{org.unique_id}", f"{org.name}")
        remaining -= limit
        limit = min(remaining, DEFAULT_PAGER_SIZE)
        console.print(table)


@api_app.command()
def simulations(
        ctx: typer.Context,
        queued: bool = typer.Option(
        False,
        "--queued",
        help="Show the current queue",
        )
):
    """List simulations"""
    simulations_api = SimulationsApi(ctx.obj.api_client)
    simulations = simulations_api.simulations_list()

    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column(
        "Id", width=5
    )  # , style="dark_green", header_style="bold dark_green")
    table.add_column("Name", width=20, justify="left")
    table.add_column("Status", justify="left")

    remaining = simulations.count
    limit = DEFAULT_PAGER_SIZE
    offset = 0
    while remaining == simulations.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        simulations = simulations_api.simulations_list(
            limit=limit, offset=offset
        )
        for i, simulation in enumerate(simulations.results):
            status = simulations_api.simulations_status_list(simulation.id)
            if status.name == "finished":
                txt = f":heavy_check_mark: [bold green]{status.name}[/bold green] "
            elif status.name == "created":
                txt = f"[dim grey27]{status.name}[/dim grey27] "
            elif status.name == "initialized":
                txt = f"[bold dark_violet]{status.name}[/bold dark_violet] "
            else:
                txt = f"[bold red]{status.name}[/bold red]"
            table.add_row(f"{simulation.id}", f"{simulation.name}", txt)
        console.print(table)
        remaining -= limit
        offset += DEFAULT_PAGER_SIZE


@api_app.command()
def settings(
        ctx: typer.Context,
        organisation: str = typer.Option(None, help="Unique-id of the organisation to set as default"),
        scenario_folder: Path = typer.Option(
            None,
            dir_okay=True,
            writable=True,
            resolve_path=True,
            help="Specify a the folder for your scenario's.",
        ),
        result_folder: Path = typer.Option(
            None,
            dir_okay=True,
            writable=True,
            resolve_path=True,
            help="Specify a results folder.",
        )
):
    """Set default settings"""
    console.rule(
        ":wrench: Configuring defaults for the 3Di cli", style="bold blue"
    )
    console.print(Padding("", (1, 0)))
    if organisation is None:
        console.rule("Please choose an organisation", style="bold blue")
        if (
            ctx.obj.organisation_uuid
            and Confirm.ask(
                f"Change current organisation {ctx.obj.organisation_uuid}?"
            )
            or not ctx.obj.organisation_uuid
        ):
            organisations(ctx)
            console.rule(":pencil2:", style="bold blue")
            organisation = Prompt.ask(
                "Set default organisation. UNIQUE_ID",
                default=ctx.obj.organisation_uuid,
            )
        else:
            organisation = ctx.obj.organisation_uuid
    if not scenario_folder:
        default_scenario_path = ctx.obj.scenario_folder or Path("./scenarios").resolve()
        print(default_scenario_path)
        scenario_folder = PathPrompt.ask(
            "Where do you keep your scenario files?",
            default=default_scenario_path
        )
    if not result_folder:
        results_folder_default = ctx.obj.result_folder or Path("./results").resolve()
        result_folder = PathPrompt.ask(
            "Where do you want to store your result files?",
            default=results_folder_default
        )
    ctx.obj.organisation_uuid = organisation
    ctx.obj.result_folder = result_folder
    ctx.obj.scenario_folder = scenario_folder
    console.print(
        f":heavy_check_mark: Default settings are saved in {ctx.obj.config_file}"
    )


@api_app.command()
def run_scenario(
        ctx: typer.Context,
        scenario: int = typer.Option(
            None,
            help="The ID of the scenario, as returned by the 'scenarios' command. "
                 "If not provided the scenarios command will be invoked so you can choose an ID"),
        model: int = typer.Option(
            None,
            help="The ID of the model you want to use, as returned by "
                 "the 'models' command. If not provided the models command will be invoked so you "
                 "can choose an ID"),
        organisation: str = typer.Option(
            None,
            help="Unique-id of the organisation to set as default"
        ),
):
    """Run a scenario"""
    if scenario is None:
        scenarios()
        scenario = IntPrompt.ask(
            "Which scenario do you want to run? ID"
        )
    if model is None:
        models(ctx)
        model = IntPrompt.ask("Which model do you want to run? ID")
    if not organisation:
        organisations(ctx)
        organisation = Prompt.ask("Which organisation do you want to use? UUID")
    scenario_to_run = ctx.obj.scenarios[scenario]["file"]
    name = ctx.obj.scenarios[scenario]["name"]
    context = {
        "threedimodel_id": model,
        "organisation_uuid": organisation,
        "simulation_name": name,
        "datetime_now": datetime.utcnow().isoformat(),
    }
    parser = ScenarioParser(scenario_to_run, context)
    try:
        scenario = parser.parse(
            ctx.obj.api_client, ctx.obj.websocket_settings
        )
    except (ResolveError, ApiModelError, LoadScenarioError) as err:
        console.print(f":collision: {err}", style="error")
        sys.exit(ExitCodes.SCENARIO_CONFIG_ERROR.value)

    console.rule(f"Loading scenario {name}", style="bold blue")
    scenario.simulation.save()
    console.print(f":link: URL: {scenario.simulation.instance.url}")

    try:
        console.rule(f"Starting scenario run...", style="bold blue")
        asyncio.run(scenario.execute())
    except KeyboardInterrupt:
        pass
    except FailedStep as err:
        console.print(f"{err}", style="error")
        sys.exit(ExitCodes.RUN_SCENARIO_ERROR.value)
    except websockets.exceptions.InvalidStatusCode as err:
        console.print(f"{err}", style="error")
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    else:
        success_panel = Panel(
            f"Run for scenario {name} successful",
            expand=True,
            box=box.DOUBLE,
            border_style="bold spring_green4",
            title=f":sparkles: Finished :sparkles:",
        )
        console.print(success_panel, justify="center")


@api_app.command()
def results(
        ctx: typer.Context,
        simulation: int = typer.Option(None, help="ID of the simulation"),
        folder: Path = typer.Option(
            None,
            dir_okay=True,
            writable=True,
            resolve_path=True,
            help="Absolute path to where the files will be stored.",
        ),
):
    """Download results of a simulation"""
    console.rule(
        ":arrow_heading_down:  Download simulation results", style="bold blue"
    )
    console.print(Padding("", (1, 0)))

    if simulation is None:
        console.rule("Please choose a simulation", style="bold blue")
        simulations(ctx)
        simulation = IntPrompt.ask("Which simulation results do you want download? ID")
    if not folder:
        result_folder = ctx.obj.result_folder
        if not result_folder:
            result_folder = Path.cwd()
        folder = PathPrompt.ask(
            "Where do you want to store the results files?",
            default=Path(f"{result_folder}/simulation-{simulation}")
        )

    try:
        folder.resolve().mkdir(parents=True)
    except FileExistsError:
        click.confirm(
            "Output folder already exists, we might override files in the folder. "
            "Do you want to continue?",
            abort=True,
        )

    simulations_api = SimulationsApi(ctx.obj.api_client)
    threedimodels_api = ThreedimodelsApi(ctx.obj.api_client)

    simulation = simulations_api.simulations_read(id=simulation)
    threedi_model_id = simulation.threedimodel_id

    gridadmin_download = threedimodels_api.threedimodels_gridadmin_download(
        threedi_model_id
    )
    f = [gridadmin_download]

    result_files = simulations_api.simulations_results_files_list(
        simulation.id
    )
    for result in result_files.results:
        if result.file.state in ["error", "removed"]:
            console.print(
                f"{result.filename} is in state {result.file.state} and will be skipped.",
                style="warning",
            )
            continue

        result_download = simulations_api.simulations_results_files_download(
            id=result.id, simulation_pk=simulation.id
        )
        f.append(result_download)
    success = download_files(f, folder)
    if success:
        console.print(
            ":heavy_check_mark: Finished downloading results", style="success"
        )
    else:
        console.print(
            ":collision: Not all files could be downloaded", style="error"
        )


@api_app.command()
def login(ctx: typer.Context):
    """Login to the specified endpoint."""
    try:
        configuration = Configuration(ctx.obj.endpoint)
        ctx.obj.credentials_prompt(configuration)
        console.print(
            f":unlock: Authenticated as {ctx.obj.username}", style="success"
        )
    except ApiException as e:
        console.print(
            f":lock: Failed to authenticate: {e.reason}", style="warning"
        )
        sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)


if __name__ == "__main__":
    app()



import typer

from threedi_cmd.commands.settings import (
    Settings,
    EndpointOption,
    get_settings,
)
from threedi_cmd.commands.models import EndpointChoices


def default_callback(
        ctx: typer.Context,
        endpoint: EndpointChoices = typer.Option(
            EndpointChoices.production, case_sensitive=False
        ),

):
    endpoint_name = EndpointOption[endpoint.value].name
    settings = get_settings(endpoint_name)
    ctx.obj = settings
    ctx.call_on_close(Settings.save_settings)


import logging
from rich.logging import RichHandler
from functools import lru_cache


@lru_cache()
def get_logger(level: str):
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)]
    )
    logger = logging.getLogger("rich")
    return logger





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


with open('README.md') as readme_file:
    readme = readme_file.read()


with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "threedi-api-client>=3.0.21",
    "PyYAML>=5.1",
    "websockets>=8.1",
    "arrow>=0.14.4,<0.14.5",  # Newer versions don't work well with 'days' directive in YAML (arrow is used in jinja2-time)
    "Jinja2>=2.10.1",
    "jinja2-time>=0.2.0",
    "slackclient>=2.8.2",
    "click>=7.1.2",
    "requests>=2.25.0",
    "rich>=9.4.0",
    "typer>=0.3.2",
]

setup_requirements = [
    "pip>=20.3.3",
    "wheel>=0.33.6",
    "flake8>=3.7.8",
    "tox>=3.14.0",
    "coverage>=4.5.4",
    "twine>=1.14.0",
    "black>=20.8b1",
]


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

test_requirements = []

main_entry_points = [
    "scenarios=threedi_cmd.commands.scenarios:scenario_app",
    "suite=threedi_cmd.commands.suite:suite_app",
    "live=threedi_cmd.commands.active_simulations:active_sims_app",
    "api=threedi_cmd.commands.api:api_app",
    "3Di_cmd=threedi_cmd.commands.main:app",
]

setup(
    author="Jelle Prins",
    author_email='info@nelen-schuurmans.nl',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
        'Topic :: Scientific/Engineering',
    ],
    description="Python 3Di command line client",
    long_description_content_type='text/markdown',
    long_description=readme + '\n\n' + history,
    install_requires=requirements,
    license="MIT license",
    entry_points={
        "console_scripts": main_entry_points
    },
    include_package_data=True,
    keywords='3Di, client, command line, scenario',
    name='threedi_cmd',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
        include=['threedi_cmd', 'threedi_cmd.*'],
    ),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nens/threedi-cmd',
    version=find_version('threedi_cmd', 'version.py'),
    zip_safe=False,
)



import sys
from openapi_client import Configuration, ApiException
from rich.table import Table
from rich import box
import typer

from threedi_cmd.commands.settings import (
    EndpointOption,
    ScenariosMeta,
    get_settings,
)
from threedi_cmd.commands.callbacks import default_callback
from threedi_cmd.console import console
from threedi_cmd.logger import get_logger

logger = get_logger("INFO")

DEFAULT_PAGER_SIZE = 20


scenario_app = typer.Typer(callback=default_callback)


@scenario_app.command()
def scenarios(ctx: typer.Context):
    """List local scenarios"""
    table = Table(show_header=True, box=box.HORIZONTALS, show_lines=True)
    table.add_column(
        "Id", width=5
    )
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )
    table.add_column("Description", justify="left", no_wrap=False)
    table.add_column(
        "Caution", justify="left", style="orange3", header_style="bold orange3"
    )
    table.add_column(
        "yaml", justify="left", style="dim blue", header_style="bold blue"
    )
    s = ScenariosMeta(ctx.obj.scenario_folder)
    for i, scenario in enumerate(s.scenarios, start=0):
        if isinstance(scenario, dict):
            s = ""
            if scenario.get("known_constraints"):
                for k, v in scenario["known_constraints"].items():
                    s += f"{k}: {v} \n"
            table.add_row(
                f"{i}",
                f"{scenario['name']}",
                f"{scenario['description']}",
                f"{s}",
                f"{scenario['file'].stem}",
            )
    console.print(table)


if __name__ == "__main__":
    scenario_app()


from functools import lru_cache

from threedi_cmd.commands.app_definitions import registry
from threedi_cmd.plugins.tools import discover


@lru_cache()
def get_app():
    plugin_registry = discover()
    registry.apps.update(plugin_registry.apps)
    for app_name, app_meta in registry.apps.items():
        if not app_meta.add_to:
            continue
        add_to_meta = registry.apps[app_meta.add_to]
        add_to_meta.app.add_typer(app_meta.app, name=app_meta.name, help=app_meta.help)
    return registry.apps["core"].app


app = get_app()

if __name__ == "__main__":
    app = get_app()
    app()



import typer

from threedi_cmd.commands.active_simulations import active_sims_app
from threedi_cmd.commands.scenarios import scenario_app
from threedi_cmd.commands.api import api_app
from threedi_cmd.commands.suite import suite_app
from threedi_cmd.plugins.models import AppRegistry
from threedi_cmd.plugins.models import AppMeta


core = typer.Typer()

# main app definition
core = AppMeta(
    app=core,
    name="core",
    help="Interact with various parts of 3Di"
)

live = AppMeta(
    app=active_sims_app,
    name="live",
    help="Get real time updates of running simulations",
    add_to="core"
)

scenarios = AppMeta(
    app=scenario_app,
    name="scenarios",
    help="Manage your local scenarios",
    add_to="core"
)

api = AppMeta(
    app=api_app,
    name="api",
    help="Interact with with the 3Di API",
    add_to="core"
)

suite = AppMeta(
    app=suite_app,
    name="suite",
    help="Run a scenario suite",
    add_to="core"
)

registry = AppRegistry(
    apps={
        inst.name: inst for inst in
        [core, live, scenarios, api]
    }
)

